@@ -45,6 +45,7 @@
 #define to_str(arg) str(arg)
 
 #define streq(s1, s2) (strcmp (s1, s2) == 0)
+#define strneq(s1, s2, n) (strncmp (s1, s2, n) == 0)
 
 #if !DEBUG
 # define xmalloc(size)          malloc_wrap(size)
@@ -117,7 +118,9 @@
 
 #define DEBUG_FILE "debug.txt"
 
-#define VERSION "0.60"
+#define MAX_ATTRIBUTE_CHARS (5 * 2)
+
+#define VERSION "0.62"
 
 typedef enum { false, true } bool;
 
@@ -194,7 +197,8 @@ static const struct {
 };
 
 enum {
-    OPT_CLEAN = 1,
+    OPT_ATTR = 1,
+    OPT_CLEAN,
     OPT_CLEAN_ALL,
     OPT_EXCLUDE_RANDOM,
     OPT_HELP,
@@ -202,6 +206,7 @@ enum {
 };
 static int opt_type;
 static const struct option long_opts[] = {
+    { "attr",           required_argument, &opt_type, OPT_ATTR           },
     { "clean",          no_argument,       &opt_type, OPT_CLEAN          },
     { "clean-all",      no_argument,       &opt_type, OPT_CLEAN_ALL      },
     { "exclude-random", required_argument, &opt_type, OPT_EXCLUDE_RANDOM },
@@ -210,6 +215,19 @@ static const struct option long_opts[] = {
     {  NULL,            0,                 NULL,      0                  },
 };
 
+enum attr_type {
+    ATTR_BOLD = 0x01,
+    ATTR_UNDERSCORE = 0x02,
+    ATTR_BLINK = 0x04,
+    ATTR_REVERSE = 0x08,
+    ATTR_CONCEALED = 0x10
+};
+struct attr {
+    const char *name;
+    unsigned int val;
+    enum attr_type type;
+};
+
 static FILE *stream;
 #if DEBUG
 static FILE *log;
@@ -221,28 +239,31 @@ static void **vars_list;
 static bool clean;
 static bool clean_all;
 
+static char attr[MAX_ATTRIBUTE_CHARS + 1];
 static char *exclude;
 
 static const char *program_name;
 
 static void process_opts (int, char **);
+static void process_opt_attr (const char *);
+static void write_attr (const struct attr *, unsigned int *);
 static void print_hint (void);
 static void print_help (void);
 static void print_version (void);
 static void cleanup (void);
 static void free_color_names (struct color_name **);
-static void process_args (unsigned int, char **, bool *, const struct color **, const char **, FILE **);
+static void process_args (unsigned int, char **, char *, const struct color **, const char **, FILE **);
 static void process_file_arg (const char *, const char **, FILE **);
 static void skip_path_colors (const char *, const char *, const struct stat *);
-static void gather_color_names (const char *, bool *, struct color_name **);
-static void read_print_stream (bool, const struct color **, const char *, FILE *);
+static void gather_color_names (const char *, char *, struct color_name **);
+static void read_print_stream (const char *, const struct color **, const char *, FILE *);
 static void merge_print_line (const char *, const char *, FILE *);
 static void complete_part_line (const char *, char **, FILE *);
 static bool get_next_char (char *, const char **, FILE *, bool *);
 static void save_char (char, char **, size_t *, size_t *);
 static void find_color_entries (struct color_name **, const struct color **);
 static void find_color_entry (const struct color_name *, unsigned int, const struct color **);
-static void print_line (bool, const struct color **, const char * const, unsigned int);
+static void print_line (const char *, const struct color **, const char * const, unsigned int);
 static void print_clean (const char *);
 static bool is_esc (const char *);
 static const char *get_end_of_esc (const char *);
@@ -250,9 +271,9 @@ static const char *get_end_of_text (const char *);
 static void print_text (const char *, size_t);
 static bool gather_esc_offsets (const char *, const char **, const char **);
 static bool validate_esc_clean_all (const char **);
-static bool validate_esc_clean (int, unsigned int, const char **, bool *);
+static bool validate_esc_clean (int, unsigned int, unsigned int *, const char **, bool *);
 static bool is_reset (int, unsigned int, const char **);
-static bool is_bold (int, unsigned int, const char **);
+static bool is_attr (int, unsigned int, unsigned int, const char **);
 static bool is_fg_color (int, const char **);
 static bool is_bg_color (int, unsigned int, const char **);
 #if !DEBUG
@@ -283,8 +304,6 @@ main (int argc, char **argv)
 {
     unsigned int arg_cnt;
 
-    bool bold = false;
-
     const struct color *colors[2] = {
         NULL, /* foreground */
         NULL, /* background */
@@ -301,6 +320,8 @@ main (int argc, char **argv)
     log = open_file (DEBUG_FILE, "w");
 #endif
 
+    attr[0] = '\0';
+
     process_opts (argc, argv);
 
     arg_cnt = argc - optind;
@@ -332,8 +353,8 @@ main (int argc, char **argv)
     if (clean || clean_all)
       process_file_arg (argv[optind], &file, &stream);
     else
-      process_args (arg_cnt, &argv[optind], &bold, colors, &file, &stream);
-    read_print_stream (bold, colors, file, stream);
+      process_args (arg_cnt, &argv[optind], &attr[0], colors, &file, &stream);
+    read_print_stream (&attr[0], colors, file, stream);
 
     RELEASE_VAR (exclude);
 
@@ -361,6 +382,9 @@ process_opts (int argc, char **argv)
             case 0: /* long opts */
               switch (opt_type)
                 {
+                  case OPT_ATTR:
+                    process_opt_attr (optarg);
+                    break;
                   case OPT_CLEAN:
                     clean = true;
                     break;
@@ -406,6 +430,70 @@ process_opts (int argc, char **argv)
       }
 }
 
+static void
+process_opt_attr (const char *p)
+{
+    /* If attributes are added to this "list", also increase MAX_ATTRIBUTE_CHARS!  */
+    const struct attr attrs[] = {
+        { "bold",       1, ATTR_BOLD       },
+        { "underscore", 4, ATTR_UNDERSCORE },
+        { "blink",      5, ATTR_BLINK      },
+        { "reverse",    7, ATTR_REVERSE    },
+        { "concealed",  8, ATTR_CONCEALED  },
+    };
+    unsigned int attr_types = 0;
+
+    while (*p)
+      {
+        const char *s;
+        if (!isalnum (*p))
+          vfprintf_fail (formats[FMT_GENERIC], "--attr switch must be provided a string");
+        s = p;
+        while (isalnum (*p))
+          p++;
+        if (*p != '\0' && *p != ',')
+          vfprintf_fail (formats[FMT_GENERIC], "--attr switch must have strings separated by ,");
+        else
+          {
+            bool valid_attr = false;
+            unsigned int i;
+            for (i = 0; i < sizeof (attrs) / sizeof (struct attr); i++)
+              {
+                const size_t name_len = strlen (attrs[i].name);
+                if ((size_t)(p - s) == name_len && strneq (s, attrs[i].name, name_len))
+                  {
+                    write_attr (&attrs[i], &attr_types);
+                    valid_attr = true;
+                    break;
+                  }
+              }
+            if (!valid_attr)
+              {
+                char *attr_invalid = xmalloc ((p - s) + 1);
+                STACK_VAR (attr_invalid);
+                strncpy (attr_invalid, s, p - s);
+                attr_invalid[p - s] = '\0';
+                vfprintf_fail ("--attr switch attribute '%s' is not valid", attr_invalid);
+              }
+          }
+        if (*p)
+          p++;
+      }
+}
+
+static void
+write_attr (const struct attr *attr_i, unsigned int *attr_types)
+{
+    const unsigned int val = attr_i->val;
+    const enum attr_type attr_type = attr_i->type;
+    const char *attr_name = attr_i->name;
+
+    if (*attr_types & attr_type)
+      vfprintf_fail ("--attr switch has attribute '%s' twice or more", attr_name);
+    snprintf (attr + strlen (attr), 3, "%u;", val);
+    *attr_types |= attr_type;
+}
+
 static void
 print_hint (void)
 {
@@ -450,13 +538,11 @@ print_help (void)
         const char *short_opt = NULL;
         unsigned int i;
         for (i = 0; i < sizeof (short_opts) / sizeof (struct short_opt); i++)
-          {
-            if (streq (opt->name, short_opts[i].name))
-              {
-                short_opt = short_opts[i].short_opt;
-                break;
-              }
-          }
+          if (streq (opt->name, short_opts[i].name))
+            {
+              short_opt = short_opts[i].short_opt;
+              break;
+            }
         if (short_opt)
           printf ("\t\t-%s, --%s\n", short_opt, opt->name);
         else
@@ -475,22 +561,23 @@ print_version (void)
 #endif
     const char *version_prefix, *version_string;
     const char *c_flags, *ld_flags, *cpp_flags;
+    const char *const desc_flags_unknown = "unknown";
     struct bytes_size bytes_size;
     bool debug;
 #ifdef CFLAGS
     c_flags = to_str (CFLAGS);
 #else
-    c_flags = "unknown";
+    c_flags = desc_flags_unknown;
 #endif
 #ifdef LDFLAGS
     ld_flags = to_str (LDFLAGS);
 #else
-    ld_flags = "unknown";
+    ld_flags = desc_flags_unknown;
 #endif
 #ifdef CPPFLAGS
     cpp_flags = to_str (CPPFLAGS);
 #else
-    cpp_flags = "unknown";
+    cpp_flags = desc_flags_unknown;
 #endif
 #if DEBUG
     debug = true;
@@ -550,7 +637,7 @@ free_color_names (struct color_name **color_names)
 }
 
 static void
-process_args (unsigned int arg_cnt, char **arg_strings, bool *bold, const struct color **colors, const char **file, FILE **stream)
+process_args (unsigned int arg_cnt, char **arg_strings, char *attr, const struct color **colors, const char **file, FILE **stream)
 {
     int ret;
     char *p;
@@ -560,7 +647,7 @@ process_args (unsigned int arg_cnt, char **arg_strings, bool *bold, const struct
     const char *color_string = arg_cnt >= 1 ? arg_strings[0] : NULL;
     const char *file_string  = arg_cnt == 2 ? arg_strings[1] : NULL;
 
-    assert (color_string);
+    assert (color_string != NULL);
 
     if (streq (color_string, "-"))
       {
@@ -587,9 +674,9 @@ process_args (unsigned int arg_cnt, char **arg_strings, bool *bold, const struct
           vfprintf_fail (formats[FMT_STRING], "one color pair allowed only for string", color_string);
       }
 
-    gather_color_names (color_string, bold, color_names);
+    gather_color_names (color_string, attr, color_names);
 
-    assert (color_names[FOREGROUND]);
+    assert (color_names[FOREGROUND] != NULL);
 
     if (color_names[BACKGROUND])
       {
@@ -605,6 +692,7 @@ process_args (unsigned int arg_cnt, char **arg_strings, bool *bold, const struct
       }
 
     find_color_entries (color_names, colors);
+    assert (colors[FOREGROUND] != NULL);
     free_color_names (color_names);
 
     if (!colors[FOREGROUND]->code && colors[BACKGROUND] && colors[BACKGROUND]->code)
@@ -650,8 +738,8 @@ process_file_arg (const char *file_string, const char **file, FILE **stream)
         *file = "stdin";
       }
 
-    assert (*stream);
-    assert (*file);
+    assert (*stream != NULL);
+    assert (*file != NULL);
 }
 
 static void
@@ -691,21 +779,21 @@ skip_path_colors (const char *color_string, const char *file_string, const struc
 
     if (have_file)
       {
-        const char *file_exists = color_string;
+        const char *file_existing = color_string;
         if (file_string)
-          vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_exists, "cannot be used as color string");
+          vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_existing, "cannot be used as color string");
         else
           {
             if (VALID_FILE_TYPE (mode))
-              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_exists, "must be preceded by color string");
+              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_existing, "must be preceded by color string");
             else
-              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_exists, "is not a valid file type");
+              vfprintf_fail (formats[FMT_QUOTE], get_file_type (mode), file_existing, "is not a valid file type");
           }
       }
 }
 
 static void
-gather_color_names (const char *color_string, bool *bold, struct color_name **color_names)
+gather_color_names (const char *color_string, char *attr, struct color_name **color_names)
 {
     unsigned int index;
     char *color, *p, *str;
@@ -725,7 +813,7 @@ gather_color_names (const char *color_string, bool *bold, struct color_name **co
           }
         else
           p = color + strlen (color);
-        assert (p);
+        assert (p != NULL);
 
         for (ch = color; *ch; ch++)
           if (!isalpha (*ch))
@@ -743,7 +831,7 @@ gather_color_names (const char *color_string, bool *bold, struct color_name **co
             switch (index)
               {
                 case FOREGROUND:
-                  *bold = true;
+                  snprintf (attr + strlen (attr), 3, "1;");
                   break;
                 case BACKGROUND:
                   vfprintf_fail (formats[FMT_COLOR], tables[BACKGROUND].desc, color, "cannot be bold");
@@ -769,7 +857,7 @@ gather_color_names (const char *color_string, bool *bold, struct color_name **co
 }
 
 static void
-read_print_stream (bool bold, const struct color **colors, const char *file, FILE *stream)
+read_print_stream (const char *attr, const struct color **colors, const char *file, FILE *stream)
 {
     char buf[BUF_SIZE + 1];
     unsigned int flags = 0;
@@ -800,13 +888,13 @@ read_print_stream (bool bold, const struct color **colors, const char *file, FIL
               vfprintf_fail (formats[FMT_FILE], file, "unrecognized line ending");
             p = eol + SKIP_LINE_ENDINGS (flags);
             *eol = '\0';
-            print_line (bold, colors, line, flags);
+            print_line (attr, colors, line, flags);
             line = p;
           }
         if (feof (stream))
           {
             if (*line != '\0')
-              print_line (bold, colors, line, 0);
+              print_line (attr, colors, line, 0);
           }
         else if (*line != '\0')
           {
@@ -814,7 +902,7 @@ read_print_stream (bool bold, const struct color **colors, const char *file, FIL
             if ((clean || clean_all) && (p = strrchr (line, '\033')))
               merge_print_line (line, p, stream);
             else
-              print_line (bold, colors, line, 0);
+              print_line (attr, colors, line, 0);
           }
       }
 }
@@ -881,7 +969,7 @@ complete_part_line (const char *p, char **buf, FILE *stream)
             if (read_from_stream)
               save_char (ch, buf, &i, &size);
           }
-        else /* read next character */
+        else /* got next character */
           {
             got_next_char = true;
             break;
@@ -1021,7 +1109,7 @@ find_color_entry (const struct color_name *color_name, unsigned int index, const
 }
 
 static void
-print_line (bool bold, const struct color **colors, const char *const line, unsigned int flags)
+print_line (const char *attr, const struct color **colors, const char *const line, unsigned int flags)
 {
     /* --clean[-all] */
     if (clean || clean_all)
@@ -1032,7 +1120,7 @@ print_line (bool bold, const struct color **colors, const char *const line, unsi
         if (colors[BACKGROUND] && colors[BACKGROUND]->code)
           printf ("\033[%s", colors[BACKGROUND]->code);
         if (colors[FOREGROUND]->code)
-          printf ("\033[%s%s%s\033[0m", bold ? "1;" : "", colors[FOREGROUND]->code, line);
+          printf ("\033[%s%s%s\033[0m", attr, colors[FOREGROUND]->code, line);
         else
           printf (formats[FMT_GENERIC], line);
       }
@@ -1116,8 +1204,9 @@ gather_esc_offsets (const char *p, const char **start, const char **end)
         else if (clean)
           {
             bool check_values;
-            unsigned int iter = 0;
+            unsigned int prev_iter, iter;
             const char *digit;
+            prev_iter = iter = 0;
             do {
               check_values = false;
               iter++;
@@ -1138,7 +1227,7 @@ gather_esc_offsets (const char *p, const char **start, const char **end)
                     val[i] = *digit++;
                   val[i] = '\0';
                   value = atoi (val);
-                  valid = validate_esc_clean (value, iter, &p, &check_values);
+                  valid = validate_esc_clean (value, iter, &prev_iter, &p, &check_values);
                 }
             } while (check_values);
           }
@@ -1163,14 +1252,15 @@ validate_esc_clean_all (const char **p)
 }
 
 static bool
-validate_esc_clean (int value, unsigned int iter, const char **p, bool *check_values)
+validate_esc_clean (int value, unsigned int iter, unsigned int *prev_iter, const char **p, bool *check_values)
 {
     if (is_reset (value, iter, p))
       return true;
-    else if (is_bold (value, iter, p))
+    else if (is_attr (value, iter, *prev_iter, p))
       {
         (*p)++;
         *check_values = true;
+        *prev_iter = iter;
         return false; /* partial escape sequence, need another valid value */
       }
     else if (is_fg_color (value, p))
@@ -1188,9 +1278,9 @@ is_reset (int value, unsigned int iter, const char **p)
 }
 
 static bool
-is_bold (int value, unsigned int iter, const char **p)
+is_attr (int value, unsigned int iter, unsigned int prev_iter, const char **p)
 {
-    return (value == 1 && iter == 1 && **p == ';');
+    return ((value > 0 && value < 10) && (iter - prev_iter == 1) && **p == ';');
 }
 
 static bool
